#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Module for tomographic reconstruction algorithms"""

__authors__ = ["P. Paleo"]
__license__ = "MIT"
__date__ = "01/08/2019"

import logging
import numpy as np

from .common import pyopencl
from .processing import OpenclProcessing
from .backprojection import Backprojection
from .projection import Projection
from .linalg import LinAlg

import pyopencl.array as parray
from pyopencl.elementwise import ElementwiseKernel
logger = logging.getLogger(__name__)

cl = pyopencl


class ReconstructionAlgorithm(OpenclProcessing):
    """
    A parent class for all iterative tomographic reconstruction algorithms

    :param sino_shape: shape of the sinogram. The sinogram is in the format
                       (n_b, n_a) where n_b is the number of detector bins and
                       n_a is the number of angles.
    :param slice_shape: Optional, shape of the reconstructed slice.
                        By default, it is a square slice where the dimension
                        is the "x dimension" of the sinogram (number of bins).
    :param axis_position: Optional, axis position. Default is `(shape[1]-1)/2.0`.
    :param angles: Optional, a list of custom angles in radian.
    :param ctx: actual working context, left to None for automatic
                initialization from device type or platformid/deviceid
    :param devicetype: type of device, can be "CPU", "GPU", "ACC" or "ALL"
    :param platformid: integer with the platform_identifier, as given by clinfo
    :param deviceid: Integer with the device identifier, as given by clinfo
    :param profile: switch on profiling to be able to profile at the kernel level,
                    store profiling elements (makes code slightly slower)
    """

    def __init__(self, sino_shape, slice_shape=None, axis_position=None, angles=None,
                 ctx=None, devicetype="all", platformid=None, deviceid=None,
                 profile=False
                 ):
        OpenclProcessing.__init__(self, ctx=ctx, devicetype=devicetype,
                                  platformid=platformid, deviceid=deviceid,
                                  profile=profile)

        # Create a backprojector
        self.backprojector = Backprojection(
            sino_shape,
            slice_shape=slice_shape,
            axis_position=axis_position,
            angles=angles,
            ctx=self.ctx,
            profile=profile
        )
        # Create a projector
        self.projector = Projection(
            self.backprojector.slice_shape,
            self.backprojector.angles,
            axis_position=axis_position,
            detector_width=self.backprojector.num_bins,
            normalize=False,
            ctx=self.ctx,
            profile=profile
        )
        self.sino_shape = sino_shape
        self.is_cpu = self.backprojector.is_cpu
        # Arrays
        self.d_data = parray.empty(self.queue, sino_shape, dtype=np.float32)
        self.d_data.fill(0.0)
        self.d_sino = parray.empty_like(self.d_data)
        self.d_sino.fill(0.0)
        self.d_x = parray.empty(self.queue,
                                self.backprojector.slice_shape,
                                dtype=np.float32)
        self.d_x.fill(0.0)
        self.d_x_old = parray.empty_like(self.d_x)
        self.d_x_old.fill(0.0)

        self.add_to_cl_mem({
                            "d_data": self.d_data,
                            "d_sino": self.d_sino,
                            "d_x": self.d_x,
                            "d_x_old": self.d_x_old,
                            })

    def proj(self, d_slice, d_sino):
        """
        Project d_slice to d_sino
        """
        self.projector.transfer_device_to_texture(d_slice.data)  #.wait()
        self.projector.projection(dst=d_sino)

    def backproj(self, d_sino, d_slice):
        """
        Backproject d_sino to d_slice
        """
        self.backprojector.transfer_device_to_texture(d_sino.data)  #.wait()
        self.backprojector.backprojection(dst=d_slice)


class SIRT(ReconstructionAlgorithm):
    """
    A class for the SIRT algorithm

    :param sino_shape: shape of the sinogram. The sinogram is in the format
                       (n_b, n_a) where n_b is the number of detector bins and
                       n_a is the number of angles.
    :param slice_shape: Optional, shape of the reconstructed slice.
                        By default, it is a square slice where the dimension is
                        the "x dimension" of the sinogram (number of bins).
    :param axis_position: Optional, axis position. Default is `(shape[1]-1)/2.0`.
    :param angles: Optional, a list of custom angles in radian.
    :param ctx: actual working context, left to None for automatic
                initialization from device type or platformid/deviceid
    :param devicetype: type of device, can be "CPU", "GPU", "ACC" or "ALL"
    :param platformid: integer with the platform_identifier, as given by clinfo
    :param deviceid: Integer with the device identifier, as given by clinfo
    :param profile: switch on profiling to be able to profile at the kernel level,
                    store profiling elements (makes code slightly slower)

    .. warning:: This is a beta version of the SIRT algorithm. Reconstruction
            fails for at least on CPU (Xeon E3-1245 v5) using the AMD opencl
            implementation.
    """

    def __init__(self, sino_shape, slice_shape=None, axis_position=None, angles=None,
                 ctx=None, devicetype="all", platformid=None, deviceid=None,
                 profile=False
                 ):

        ReconstructionAlgorithm.__init__(self, sino_shape, slice_shape=slice_shape,
                                         axis_position=axis_position, angles=angles,
                                         ctx=ctx, devicetype=devicetype, platformid=platformid,
                                         deviceid=deviceid, profile=profile)
        self.compute_preconditioners()

    def compute_preconditioners(self):
        """
        Create a diagonal preconditioner for the projection and backprojection
        operator.
        Each term of the diagonal is the sum of the projector/backprojector
        along rows [1], i.e the projection/backprojection of an array of ones.

        [1] Jens Gregor and Thomas Benson,
            Computational Analysis and Improvement of SIRT,
            IEEE transactions on medical imaging, vol. 27, no. 7,  2008
        """

        # r_{i,i} = 1/(sum_j a_{i,j})
        slice_ones = np.ones(self.backprojector.slice_shape, dtype=np.float32)
        R = 1./self.projector.projection(slice_ones)  # could be all done on GPU, but I want extra checks
        R[np.logical_not(np.isfinite(R))] = 1.  # In the case where the rotation axis is excentred
        self.d_R = parray.to_device(self.queue, R)
        # c_{j,j} = 1/(sum_i a_{i,j})
        sino_ones = np.ones(self.sino_shape, dtype=np.float32)
        C = 1./self.backprojector.backprojection(sino_ones)
        C[np.logical_not(np.isfinite(C))] = 1.  # In the case where the rotation axis is excentred
        self.d_C = parray.to_device(self.queue, C)

        self.add_to_cl_mem({
            "d_R": self.d_R,
            "d_C": self.d_C
        })

    # TODO: compute and possibly return the residual
    def run(self, data, n_it):
        """
        Run n_it iterations of the SIRT algorithm.
        """
        cl.enqueue_copy(self.queue, self.d_data.data, np.ascontiguousarray(data.astype(np.float32)))

        d_x_old = self.d_x_old
        d_x = self.d_x
        d_R = self.d_R
        d_C = self.d_C
        d_sino = self.d_sino
        d_x *= 0

        for k in range(n_it):
            d_x_old[:] = d_x[:]
            # x{k+1} = x{k} - C A^T R (A x{k} - b)
            self.proj(d_x, d_sino)
            d_sino -= self.d_data
            d_sino *= d_R
            if self.is_cpu:
                # This sync is necessary when using CPU, while it is not for GPU
                d_sino.finish()
            self.backproj(d_sino, d_x)
            d_x *= -d_C
            d_x += d_x_old
            if self.is_cpu:
                # This sync is necessary when using CPU, while it is not for GPU
                d_x.finish()

        return d_x

    __call__ = run


class TV(ReconstructionAlgorithm):
    """
    A class for reconstruction with Total Variation regularization using the
    Chambolle-Pock TV reconstruction algorithm.

    :param sino_shape: shape of the sinogram. The sinogram is in the format
                       (n_b, n_a) where n_b is the number of detector bins and
                       n_a is the number of angles.
    :param slice_shape: Optional, shape of the reconstructed slice. By default,
                        it is a square slice where the dimension is the
                        "x dimension" of the sinogram (number of bins).
    :param axis_position: Optional, axis position. Default is
                          `(shape[1]-1)/2.0`.
    :param angles: Optional, a list of custom angles in radian.
    :param ctx: actual working context, left to None for automatic
                initialization from device type or platformid/deviceid
    :param devicetype: type of device, can be "CPU", "GPU", "ACC" or "ALL"
    :param platformid: integer with the platform_identifier, as given by clinfo
    :param deviceid: Integer with the device identifier, as given by clinfo
    :param profile: switch on profiling to be able to profile at the kernel
                    level, store profiling elements (makes code slightly slower)

    .. warning:: This is a beta version of the Chambolle-Pock TV algorithm.
            Reconstruction fails for at least on CPU (Xeon E3-1245 v5) using
            the AMD opencl implementation.
    """

    def __init__(self, sino_shape, slice_shape=None, axis_position=None, angles=None,
                 ctx=None, devicetype="all", platformid=None, deviceid=None,
                 profile=False
                 ):
        ReconstructionAlgorithm.__init__(self, sino_shape, slice_shape=slice_shape,
                                         axis_position=axis_position, angles=angles,
                                         ctx=ctx, devicetype=devicetype, platformid=platformid,
                                         deviceid=deviceid, profile=profile)
        self.compute_preconditioners()

        # Create a LinAlg instance
        self.linalg = LinAlg(self.backprojector.slice_shape, ctx=self.ctx)
        # Positivity constraint
        self.elwise_clamp = ElementwiseKernel(self.ctx, "float *a", "a[i] = max(a[i], 0.0f);")
        # Projection onto the L-infinity ball of radius Lambda
        self.elwise_proj_linf = ElementwiseKernel(
            self.ctx,
            "float2* a, float Lambda",
            "a[i].x = copysign(min(fabs(a[i].x), Lambda), a[i].x); a[i].y = copysign(min(fabs(a[i].y), Lambda), a[i].y);",
            "elwise_proj_linf"
        )
        # Additional arrays
        self.linalg.gradient(self.d_x)
        self.d_p = parray.empty_like(self.linalg.cl_mem["d_gradient"])
        self.d_q = parray.empty_like(self.d_data)
        self.d_g = self.linalg.d_image
        self.d_tmp = parray.empty_like(self.d_x)
        self.d_p.fill(0)
        self.d_q.fill(0)
        self.d_tmp.fill(0)
        self.add_to_cl_mem({
            "d_p": self.d_p,
            "d_q": self.d_q,
            "d_tmp": self.d_tmp,
        })

        self.theta = 1.0

    def compute_preconditioners(self):
        """
        Create a diagonal preconditioner for the projection and backprojection
        operator.
        Each term of the diagonal is the sum of the projector/backprojector
        along rows [2],
        i.e the projection/backprojection of an array of ones.

        [2] T. Pock, A. Chambolle,
            Diagonal preconditioning for first order primal-dual algorithms in
            convex optimization,
            International Conference on Computer Vision, 2011
        """

        # Compute the diagonal preconditioner "Sigma"
        slice_ones = np.ones(self.backprojector.slice_shape, dtype=np.float32)
        Sigma_k = 1./self.projector.projection(slice_ones)
        Sigma_k[np.logical_not(np.isfinite(Sigma_k))] = 1.
        self.d_Sigma_k = parray.to_device(self.queue, Sigma_k)
        self.d_Sigma_kp1 = self.d_Sigma_k + 1  # TODO: memory vs computation
        self.Sigma_grad = 1/2.0  # For discrete gradient, sum|D_i,j| = 2 along lines or cols

        # Compute the diagonal preconditioner "Tau"
        sino_ones = np.ones(self.sino_shape, dtype=np.float32)
        C = self.backprojector.backprojection(sino_ones)
        Tau = 1./(C + 2.)
        self.d_Tau = parray.to_device(self.queue, Tau)

        self.add_to_cl_mem({
            "d_Sigma_k": self.d_Sigma_k,
            "d_Sigma_kp1": self.d_Sigma_kp1,
            "d_Tau": self.d_Tau
        })

    def run(self, data, n_it, Lambda, pos_constraint=False):
        """
        Run n_it iterations of the TV-regularized reconstruction,
        with the regularization parameter Lambda.
        """
        cl.enqueue_copy(self.queue, self.d_data.data, np.ascontiguousarray(data.astype(np.float32)))

        d_x = self.d_x
        d_x_old = self.d_x_old
        d_tmp = self.d_tmp
        d_sino = self.d_sino
        d_p = self.d_p
        d_q = self.d_q
        d_g = self.d_g

        d_x *= 0
        d_p *= 0
        d_q *= 0

        for k in range(0, n_it):
            # Update primal variables
            d_x_old[:] = d_x[:]
            #~ x = x + Tau*div(p) - Tau*Kadj(q)
            self.backproj(d_q, d_tmp)
            self.linalg.divergence(d_p)
            # TODO: this in less than three ops (one kernel ?)
            d_g -= d_tmp  # d_g -> L.d_image
            d_g *= self.d_Tau
            d_x += d_g

            if pos_constraint:
                self.elwise_clamp(d_x)

            # Update dual variables
            #~ p = proj_linf(p + Sigma_grad*gradient(x + theta*(x - x_old)), Lambda)
            d_tmp[:] = d_x[:]
            # FIXME: mul_add is out of place, put an equivalent thing in linalg...
            #~ d_tmp.mul_add(1 + theta, d_x_old, -theta)
            d_tmp *= 1+self.theta
            d_tmp -= self.theta*d_x_old
            self.linalg.gradient(d_tmp)
            # TODO: out of place mul_add
            #~ d_p.mul_add(1, L.cl_mem["d_gradient"], Sigma_grad)
            self.linalg.cl_mem["d_gradient"] *= self.Sigma_grad
            d_p += self.linalg.cl_mem["d_gradient"]
            self.elwise_proj_linf(d_p, Lambda)

            #~ q = (q + Sigma_k*K(x + theta*(x - x_old)) - Sigma_k*data)/(1.0 + Sigma_k)
            self.proj(d_tmp, d_sino)
            # TODO: this in less instructions
            d_sino -= self.d_data
            d_sino *= self.d_Sigma_k
            d_q += d_sino
            d_q /= self.d_Sigma_kp1
        return d_x

    __call__ = run



# /*##########################################################################
#
# Copyright (c) 2017-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This script illustrates the update of a :mod:`silx.gui.plot` widget from a thread.

The problem is that plot and GUI methods should be called from the main thread.
To safely update the plot from another thread, one need to execute the update
asynchronously in the main thread.
In this example, this is achieved with
:func:`~silx.gui.utils.concurrent.submitToQtMainThread`.

In this example a thread calls submitToQtMainThread to update the curve
of a plot.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "05/09/2017"


import threading
import time

import numpy

from silx.gui import qt
from silx.gui.utils import concurrent

from silx.gui.plot import Plot1D


class UpdateThread(threading.Thread):
    """Thread updating the curve of a :class:`~silx.gui.plot.Plot1D`

    :param plot1d: The Plot1D to update."""

    def __init__(self, plot1d):
        self.plot1d = plot1d
        self.running = False
        super(UpdateThread, self).__init__()

    def start(self):
        """Start the update thread"""
        self.running = True
        super(UpdateThread, self).start()

    def run(self):
        """Method implementing thread loop that updates the plot"""
        while self.running:
            time.sleep(1)
            # Run plot update asynchronously
            concurrent.submitToQtMainThread(
                self.plot1d.addCurve,
                numpy.arange(1000),
                numpy.random.random(1000),
                resetzoom=False)

    def stop(self):
        """Stop the update thread"""
        self.running = False
        self.join(2)


def main():
    global app
    app = qt.QApplication([])

    # Create a Plot1D, set its limits and display it
    plot1d = Plot1D()
    plot1d.setLimits(0., 1000., 0., 1.)
    plot1d.show()

    # Create the thread that calls submitToQtMainThread
    updateThread = UpdateThread(plot1d)
    updateThread.start()  # Start updating the plot

    app.exec()

    updateThread.stop()  # Stop updating the plot


if __name__ == '__main__':
    main()



# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

"""
Store in a Cython module if it was compiled with OpenMP

You have to patch the setup module like that:

.. code-block:: python

    silx_include = os.path.join(top_path, "src", ""silx", "utils", "include")
    config.add_extension('my_extension',
                         include_dirs=[silx_include],
                         ...)

Then you can include it like that in your Cython module:

.. code-block:: python

    include "../../utils/_have_openmp.pxi"

"""


cdef extern from "silx_store_openmp.h":
    int COMPILED_WITH_OPENMP
_COMPILED_WITH_OPENMP = COMPILED_WITH_OPENMP



#
#    Project: Azimuthal integration
#             https://github.com/silx-kit/silx
#
#    Copyright (C) 2015-2019 European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Provide convenient URL for silx-kit projects."""

__author__ = "Valentin Valls"
__contact__ = "valentin.valls@ESRF.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "15/01/2019"


from ... import _version as version

BASE_DOC_URL = None
"""This could be patched by project packagers."""

_DEFAULT_BASE_DOC_URL = "http://www.silx.org/pub/doc/silx/{silx_doc_version}/{subpath}"
"""Identify the base URL of the project documentation.

It supportes string replacement:

- `{major}` the major version
- `{minor}` the minor version
- `{micro}` the micro version
- `{relev}` the status of the version (dev, final, rc).
- `{silx_doc_version}` is used to map the documentation stored at www.silx.org
- `{subpath}` is the subpart of the URL pointing to a specific page of the
    documentation. It is mandatory.
"""


def getDocumentationUrl(subpath):
    """Returns the URL to the documentation"""

    if version.RELEV == "final":
        # Released verison will point to a specific documentation
        silx_doc_version = "%d.%d.%d" % (version.MAJOR, version.MINOR, version.MICRO)
    else:
        # Dev versions will point to a single 'dev' documentation
        silx_doc_version = "dev"

    keyworks = {
        "silx_doc_version": silx_doc_version,
        "major": version.MAJOR,
        "minor": version.MINOR,
        "micro": version.MICRO,
        "relev": version.RELEV,
        "subpath": subpath}
    template = BASE_DOC_URL
    if template is None:
        template = _DEFAULT_BASE_DOC_URL
    return template.format(**keyworks)



# /*##########################################################################
#
# Copyright (c) 2015-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""QWidget displaying a 2D image with histograms on its sides.

The :class:`ImageView` implements this widget, and
:class:`ImageViewMainWindow` provides a main window with additional toolbar
and status bar.

Basic usage of :class:`ImageView` is through the following methods:

- :meth:`ImageView.getColormap`, :meth:`ImageView.setColormap` to update the
  default colormap to use and update the currently displayed image.
- :meth:`ImageView.setImage` to update the displayed image.

For an example of use, see `imageview.py` in :ref:`sample-code`.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "26/04/2018"


import logging
import numpy
import collections
from typing import Union

import silx
from .. import qt
from .. import colors
from .. import icons

from . import items, PlotWindow, PlotWidget, actions
from ..colors import Colormap
from ..colors import cursorColorForColormap
from .tools import LimitsToolBar
from .Profile import ProfileToolBar
from ...utils.proxy import docstring
from ...utils.deprecation import deprecated
from ...utils.enum import Enum
from .tools.RadarView import RadarView
from .utils.axis import SyncAxes
from ..utils import blockSignals
from . import _utils
from .tools.profile import manager
from .tools.profile import rois
from .actions import PlotAction

_logger = logging.getLogger(__name__)


ProfileSumResult = collections.namedtuple("ProfileResult",
                                          ["dataXRange", "dataYRange",
                                           'histoH', 'histoHRange',
                                           'histoV', 'histoVRange',
                                           "xCoords", "xData",
                                           "yCoords", "yData"])


def computeProfileSumOnRange(imageItem, xRange, yRange, cache=None):
    """
    Compute a full vertical and horizontal profile on an image item using a
    a range in the plot referential.

    Optionally takes a previous computed result to be able to skip the
    computation.

    :rtype: ProfileSumResult
    """
    data = imageItem.getValueData(copy=False)
    origin = imageItem.getOrigin()
    scale = imageItem.getScale()
    height, width = data.shape

    xMin, xMax = xRange
    yMin, yMax = yRange

    # Convert plot area limits to image coordinates
    # and work in image coordinates (i.e., in pixels)
    xMin = int((xMin - origin[0]) / scale[0])
    xMax = int((xMax - origin[0]) / scale[0])
    yMin = int((yMin - origin[1]) / scale[1])
    yMax = int((yMax - origin[1]) / scale[1])

    if (xMin >= width or xMax < 0 or
            yMin >= height or yMax < 0):
        return None

    # The image is at least partly in the plot area
    # Get the visible bounds in image coords (i.e., in pixels)
    subsetXMin = 0 if xMin < 0 else xMin
    subsetXMax = (width if xMax >= width else xMax) + 1
    subsetYMin = 0 if yMin < 0 else yMin
    subsetYMax = (height if yMax >= height else yMax) + 1

    if cache is not None:
        if ((subsetXMin, subsetXMax) == cache.dataXRange and
                (subsetYMin, subsetYMax) == cache.dataYRange):
            # The visible area of data is the same
            return cache

    # Rebuild histograms for visible area
    visibleData = data[subsetYMin:subsetYMax,
                       subsetXMin:subsetXMax]
    histoHVisibleData = numpy.nansum(visibleData, axis=0)
    histoVVisibleData = numpy.nansum(visibleData, axis=1)
    histoHMin = numpy.nanmin(histoHVisibleData)
    histoHMax = numpy.nanmax(histoHVisibleData)
    histoVMin = numpy.nanmin(histoVVisibleData)
    histoVMax = numpy.nanmax(histoVVisibleData)

    # Convert to histogram curve and update plots
    # Taking into account origin and scale
    coords = numpy.arange(2 * histoHVisibleData.size)
    xCoords = (coords + 1) // 2 + subsetXMin
    xCoords = origin[0] + scale[0] * xCoords
    xData = numpy.take(histoHVisibleData, coords // 2)
    coords = numpy.arange(2 * histoVVisibleData.size)
    yCoords = (coords + 1) // 2 + subsetYMin
    yCoords = origin[1] + scale[1] * yCoords
    yData = numpy.take(histoVVisibleData, coords // 2)

    result = ProfileSumResult(
        dataXRange=(subsetXMin, subsetXMax),
        dataYRange=(subsetYMin, subsetYMax),
        histoH=histoHVisibleData,
        histoHRange=(histoHMin, histoHMax),
        histoV=histoVVisibleData,
        histoVRange=(histoVMin, histoVMax),
        xCoords=xCoords,
        xData=xData,
        yCoords=yCoords,
        yData=yData)

    return result


class _SideHistogram(PlotWidget):
    """
    Widget displaying one of the side profile of the ImageView.

    Implement ProfileWindow
    """

    sigClose = qt.Signal()

    sigMouseMoved = qt.Signal(float, float)

    def __init__(self, parent=None, backend=None, direction=qt.Qt.Horizontal):
        super(_SideHistogram, self).__init__(parent=parent, backend=backend)
        self._direction = direction
        self.sigPlotSignal.connect(self._plotEvents)
        self._color = "blue"
        self.__profile = None
        self.__profileSum = None

    def _plotEvents(self, eventDict):
        """Callback for horizontal histogram plot events."""
        if eventDict['event'] == 'mouseMoved':
            self.sigMouseMoved.emit(eventDict['x'], eventDict['y'])

    def setProfileColor(self, color):
        self._color = color

    def setProfileSum(self, result):
        self.__profileSum = result
        if self.__profile is None:
            self.__drawProfileSum()

    def prepareWidget(self, roi):
        """Implements `ProfileWindow`"""
        pass

    def setRoiProfile(self, roi):
        """Implements `ProfileWindow`"""
        if roi is None:
            return
        self._roiColor = colors.rgba(roi.getColor())

    def getProfile(self):
        """Implements `ProfileWindow`"""
        return self.__profile

    def setProfile(self, data):
        """Implements `ProfileWindow`"""
        self.__profile = data
        if data is None:
            self.__drawProfileSum()
        else:
            self.__drawProfile()

    def __drawProfileSum(self):
        """Only draw the profile sum on the plot.

        Other elements are removed
        """
        profileSum = self.__profileSum

        try:
            self.removeCurve('profile')
        except Exception:
            pass

        if profileSum is None:
            try:
                self.removeCurve('profilesum')
            except Exception:
                pass
            return

        if self._direction == qt.Qt.Horizontal:
            xx, yy = profileSum.xCoords, profileSum.xData
        elif self._direction == qt.Qt.Vertical:
            xx, yy = profileSum.yData, profileSum.yCoords
        else:
            assert False

        self.addCurve(xx, yy,
                      xlabel='', ylabel='',
                      legend="profilesum",
                      color=self._color,
                      linestyle='-',
                      selectable=False,
                      resetzoom=False)

        self.__updateLimits()

    def __drawProfile(self):
        """Only draw the profile on the plot.

        Other elements are removed
        """
        profile = self.__profile

        try:
            self.removeCurve('profilesum')
        except Exception:
            pass

        if profile is None:
            try:
                self.removeCurve('profile')
            except Exception:
                pass
            self.setProfileSum(self.__profileSum)
            return

        if self._direction == qt.Qt.Horizontal:
            xx, yy = profile.coords, profile.profile
        elif self._direction == qt.Qt.Vertical:
            xx, yy = profile.profile, profile.coords
        else:
            assert False

        self.addCurve(xx,
                      yy,
                      legend="profile",
                      color=self._roiColor,
                      resetzoom=False)

        self.__updateLimits()

    def __updateLimits(self):
        if self.__profile:
            data = self.__profile.profile
            vMin = numpy.nanmin(data)
            vMax = numpy.nanmax(data)
        elif self.__profileSum is not None:
            if self._direction == qt.Qt.Horizontal:
                vMin, vMax = self.__profileSum.histoHRange
            elif self._direction == qt.Qt.Vertical:
                vMin, vMax = self.__profileSum.histoVRange
            else:
                assert False
        else:
            vMin, vMax = 0, 0

        # Tune the result using the data margins
        margins = self.getDataMargins()
        if self._direction == qt.Qt.Horizontal:
            _, _, vMin, vMax = _utils.addMarginsToLimits(margins, False, False, 0, 0, vMin, vMax)
        elif self._direction == qt.Qt.Vertical:
            vMin, vMax, _, _ = _utils.addMarginsToLimits(margins, False, False, vMin, vMax, 0, 0)
        else:
            assert False

        if self._direction == qt.Qt.Horizontal:
            dataAxis = self.getYAxis()
        elif self._direction == qt.Qt.Vertical:
            dataAxis = self.getXAxis()
        else:
            assert False

        with blockSignals(dataAxis):
            dataAxis.setLimits(vMin, vMax)


class ShowSideHistogramsAction(PlotAction):
    """QAction to change visibility of side histogram of a :class:`.ImageView`.

    :param plot: :class:`.ImageView` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        super(ShowSideHistogramsAction, self).__init__(
            plot, icon='side-histograms', text='Show/hide side histograms',
            tooltip='Show/hide side histogram',
            triggered=self._actionTriggered,
            checkable=True, parent=parent)

    def _actionTriggered(self, checked=False):
        if self.plot.isSideHistogramDisplayed() != checked:
            self.plot.setSideHistogramDisplayed(checked)


class AggregationModeAction(qt.QWidgetAction):
    """Action providing few filters to the image"""

    sigAggregationModeChanged = qt.Signal()

    def __init__(self, parent):
        qt.QWidgetAction.__init__(self, parent)

        toolButton = qt.QToolButton(parent)

        filterAction = qt.QAction(self)
        filterAction.setText("No filter")
        filterAction.setCheckable(True)
        filterAction.setChecked(True)
        filterAction.setProperty("aggregation", items.ImageDataAggregated.Aggregation.NONE)
        densityNoFilterAction = filterAction

        filterAction = qt.QAction(self)
        filterAction.setText("Max filter")
        filterAction.setCheckable(True)
        filterAction.setProperty("aggregation", items.ImageDataAggregated.Aggregation.MAX)
        densityMaxFilterAction = filterAction

        filterAction = qt.QAction(self)
        filterAction.setText("Mean filter")
        filterAction.setCheckable(True)
        filterAction.setProperty("aggregation", items.ImageDataAggregated.Aggregation.MEAN)
        densityMeanFilterAction = filterAction

        filterAction = qt.QAction(self)
        filterAction.setText("Min filter")
        filterAction.setCheckable(True)
        filterAction.setProperty("aggregation", items.ImageDataAggregated.Aggregation.MIN)
        densityMinFilterAction = filterAction

        densityGroup = qt.QActionGroup(self)
        densityGroup.setExclusive(True)
        densityGroup.addAction(densityNoFilterAction)
        densityGroup.addAction(densityMaxFilterAction)
        densityGroup.addAction(densityMeanFilterAction)
        densityGroup.addAction(densityMinFilterAction)
        densityGroup.triggered.connect(self._aggregationModeChanged)
        self.__densityGroup = densityGroup

        filterMenu = qt.QMenu(toolButton)
        filterMenu.addAction(densityNoFilterAction)
        filterMenu.addAction(densityMaxFilterAction)
        filterMenu.addAction(densityMeanFilterAction)
        filterMenu.addAction(densityMinFilterAction)

        toolButton.setPopupMode(qt.QToolButton.InstantPopup)
        toolButton.setMenu(filterMenu)
        toolButton.setText("Data filters")
        toolButton.setToolTip("Enable/disable filter on the image")
        icon = icons.getQIcon("aggregation-mode")
        toolButton.setIcon(icon)
        toolButton.setText("Pixel aggregation filter")

        self.setDefaultWidget(toolButton)

    def _aggregationModeChanged(self):
        self.sigAggregationModeChanged.emit()

    def setAggregationMode(self, mode):
        """Set an Aggregated enum from ImageDataAggregated"""
        for a in self.__densityGroup.actions():
            if a.property("aggregation") is mode:
                a.setChecked(True)

    def getAggregationMode(self):
        """Returns an Aggregated enum from ImageDataAggregated"""
        densityAction = self.__densityGroup.checkedAction()
        if densityAction is None:
            return items.ImageDataAggregated.Aggregation.NONE
        return densityAction.property("aggregation")


class ImageView(PlotWindow):
    """Display a single image with horizontal and vertical histograms.

    Use :meth:`setImage` to control the displayed image.
    This class also provides the :class:`silx.gui.plot.Plot` API.

    The :class:`ImageView` inherits from :class:`.PlotWindow` (which provides
    the toolbars) and also exposes :class:`.PlotWidget` API for further
    plot control (plot title, axes labels, aspect ratio, ...).

    :param parent: The parent of this widget or None.
    :param backend: The backend to use for the plot (default: matplotlib).
                    See :class:`.PlotWidget` for the list of supported backend.
    :type backend: str or :class:`BackendBase.BackendBase`
    """

    HISTOGRAMS_COLOR = 'blue'
    """Color to use for the side histograms."""

    HISTOGRAMS_HEIGHT = 200
    """Height in pixels of the side histograms."""

    IMAGE_MIN_SIZE = 200
    """Minimum size in pixels of the image area."""

    # Qt signals
    valueChanged = qt.Signal(float, float, object)
    """Signals that the data value under the cursor has changed.

    It provides: row, column, data value.

    When the cursor is over an histogram, either row or column is Nan
    and the provided data value is the histogram value
    (i.e., the sum along the corresponding row/column).
    Row and columns are either Nan or integer values.
    """

    class ProfileWindowBehavior(Enum):
        """ImageView's profile window behavior options"""

        POPUP = 'popup'
        """All profiles are displayed in pop-up windows"""

        EMBEDDED = 'embedded'
        """Horizontal, vertical and cross profiles are displayed in
        sides widgets, others are displayed in pop-up windows.
        """

    def __init__(self, parent=None, backend=None):
        self._imageLegend = '__ImageView__image' + str(id(self))
        self._cache = None  # Store currently visible data information

        super(ImageView, self).__init__(parent=parent, backend=backend,
                                        resetzoom=True, autoScale=False,
                                        logScale=False, grid=False,
                                        curveStyle=False, colormap=True,
                                        aspectRatio=True, yInverted=True,
                                        copy=True, save=True, print_=True,
                                        control=False, position=False,
                                        roi=False, mask=True)

        # Enable mask synchronisation to use it in profiles
        maskToolsWidget = self.getMaskToolsDockWidget().widget()
        maskToolsWidget.setItemMaskUpdated(True)

        self.__showSideHistogramsAction = ShowSideHistogramsAction(self, self)
        self.__showSideHistogramsAction.setChecked(True)

        self.__aggregationModeAction = AggregationModeAction(self)
        self.__aggregationModeAction.sigAggregationModeChanged.connect(self._aggregationModeChanged)

        if parent is None:
            self.setWindowTitle('ImageView')

        if silx.config.DEFAULT_PLOT_IMAGE_Y_AXIS_ORIENTATION == 'downward':
            self.getYAxis().setInverted(True)

        self._initWidgets(backend)

        toolBar = self.toolBar()
        toolBar.addAction(self.__showSideHistogramsAction)
        toolBar.addAction(self.__aggregationModeAction)

        self.__profileWindowBehavior = self.ProfileWindowBehavior.POPUP
        self.__profile = ProfileToolBar(plot=self)
        self.addToolBar(self.__profile)

    def _initWidgets(self, backend):
        """Set-up layout and plots."""
        self._histoHPlot = _SideHistogram(backend=backend, parent=self, direction=qt.Qt.Horizontal)
        widgetHandle = self._histoHPlot.getWidgetHandle()
        widgetHandle.setMinimumHeight(self.HISTOGRAMS_HEIGHT)
        widgetHandle.setMaximumHeight(self.HISTOGRAMS_HEIGHT)
        self._histoHPlot.setInteractiveMode('zoom')
        self._histoHPlot.setDataMargins(0., 0., 0.1, 0.1)
        self._histoHPlot.sigMouseMoved.connect(self._mouseMovedOnHistoH)
        self._histoHPlot.setProfileColor(self.HISTOGRAMS_COLOR)

        self._histoVPlot = _SideHistogram(backend=backend, parent=self, direction=qt.Qt.Vertical)
        widgetHandle = self._histoVPlot.getWidgetHandle()
        widgetHandle.setMinimumWidth(self.HISTOGRAMS_HEIGHT)
        widgetHandle.setMaximumWidth(self.HISTOGRAMS_HEIGHT)
        self._histoVPlot.setInteractiveMode('zoom')
        self._histoVPlot.setDataMargins(0.1, 0.1, 0., 0.)
        self._histoVPlot.sigMouseMoved.connect(self._mouseMovedOnHistoV)
        self._histoVPlot.setProfileColor(self.HISTOGRAMS_COLOR)

        self.setPanWithArrowKeys(True)
        self.setInteractiveMode('zoom')  # Color set in setColormap
        self.sigPlotSignal.connect(self._imagePlotCB)
        self.sigActiveImageChanged.connect(self._activeImageChangedSlot)

        self._radarView = RadarView(parent=self)
        self._radarView.setPlotWidget(self)

        self.__syncXAxis = SyncAxes([self.getXAxis(), self._histoHPlot.getXAxis()])
        self.__syncYAxis = SyncAxes([self.getYAxis(), self._histoVPlot.getYAxis()])

        self.__setCentralWidget()

    def __setCentralWidget(self):
        """Set central widget with all its content"""
        layout = qt.QGridLayout()
        layout.addWidget(self.getWidgetHandle(), 0, 0)
        layout.addWidget(self._histoVPlot, 0, 1)
        layout.addWidget(self._histoHPlot, 1, 0)
        layout.addWidget(self._radarView, 1, 1, 1, 2)
        layout.addWidget(self.getColorBarWidget(), 0, 2)

        self._radarView.setMinimumWidth(self.IMAGE_MIN_SIZE)
        self._radarView.setMinimumHeight(self.HISTOGRAMS_HEIGHT)
        self._histoHPlot.setMinimumWidth(self.IMAGE_MIN_SIZE)
        self._histoVPlot.setMinimumHeight(self.HISTOGRAMS_HEIGHT)

        layout.setColumnStretch(0, 1)
        layout.setColumnStretch(1, 0)
        layout.setRowStretch(0, 1)
        layout.setRowStretch(1, 0)

        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        centralWidget = qt.QWidget(self)
        centralWidget.setLayout(layout)
        self.setCentralWidget(centralWidget)

    @docstring(PlotWidget)
    def setBackend(self, backend):
        # Use PlotWidget here since we override PlotWindow behavior
        PlotWidget.setBackend(self, backend)
        self.__setCentralWidget()

    def _dirtyCache(self):
        self._cache = None

    def getAggregationModeAction(self):
        return self.__aggregationModeAction

    def _aggregationModeChanged(self):
        item = self._getItem("image", self._imageLegend)
        if item is None:
            return
        aggregationMode = self.__aggregationModeAction.getAggregationMode()
        if aggregationMode is not None and isinstance(item, items.ImageDataAggregated):
            item.setAggregationMode(aggregationMode)
        else:
            # It means the item type have to be changed
            self.removeImage(self._imageLegend)
            image = item.getData(copy=False)
            if image is None:
                return
            origin = item.getOrigin()
            scale = item.getScale()
            self.setImage(image, origin, scale, copy=False, resetzoom=False)

    def getShowSideHistogramsAction(self):
        return self.__showSideHistogramsAction

    def setSideHistogramDisplayed(self, show):
        """Display or not the side histograms"""
        if self.isSideHistogramDisplayed() == show:
            return
        self._histoHPlot.setVisible(show)
        self._histoVPlot.setVisible(show)
        self._radarView.setVisible(show)
        self.__showSideHistogramsAction.setChecked(show)
        if show:
            # Probably have to be computed
            self._updateHistograms()

    def isSideHistogramDisplayed(self):
        """True if the side histograms are displayed"""
        return self._histoHPlot.isVisible()

    def _updateHistograms(self):
        """Update histograms content using current active image."""
        if not self.isSideHistogramDisplayed():
            # The histogram computation can be skipped
            return

        activeImage = self.getActiveImage()
        if activeImage is not None:
            xRange = self.getXAxis().getLimits()
            yRange = self.getYAxis().getLimits()
            result = computeProfileSumOnRange(activeImage, xRange, yRange, self._cache)
            self._cache = result
            self._histoHPlot.setProfileSum(result)
            self._histoVPlot.setProfileSum(result)

    # Plots event listeners

    def _imagePlotCB(self, eventDict):
        """Callback for imageView plot events."""
        if eventDict['event'] == 'mouseMoved':
            activeImage = self.getActiveImage()
            if activeImage is not None:
                data = activeImage.getData(copy=False)
                height, width = data.shape[0:2]

                # Get corresponding coordinate in image
                origin = activeImage.getOrigin()
                scale = activeImage.getScale()
                if (eventDict['x'] >= origin[0] and
                        eventDict['y'] >= origin[1]):
                    x = int((eventDict['x'] - origin[0]) / scale[0])
                    y = int((eventDict['y'] - origin[1]) / scale[1])

                    if x >= 0 and x < width and y >= 0 and y < height:
                        self.valueChanged.emit(float(x), float(y),
                                               data[y][x])

        elif eventDict['event'] == 'limitsChanged':
            self._updateHistograms()

    def _mouseMovedOnHistoH(self, x, y):
        if self._cache is None:
            return
        activeImage = self.getActiveImage()
        if activeImage is None:
            return

        xOrigin = activeImage.getOrigin()[0]
        xScale = activeImage.getScale()[0]

        minValue = xOrigin + xScale * self._cache.dataXRange[0]

        if x >= minValue:
            data = self._cache.histoH
            column = int((x - minValue) / xScale)
            if column >= 0 and column < data.shape[0]:
                self.valueChanged.emit(
                    float('nan'),
                    float(column + self._cache.dataXRange[0]),
                    data[column])

    def _mouseMovedOnHistoV(self, x, y):
        if self._cache is None:
            return
        activeImage = self.getActiveImage()
        if activeImage is None:
            return

        yOrigin = activeImage.getOrigin()[1]
        yScale = activeImage.getScale()[1]

        minValue = yOrigin + yScale * self._cache.dataYRange[0]

        if y >= minValue:
            data = self._cache.histoV
            row = int((y - minValue) / yScale)
            if row >= 0 and row < data.shape[0]:
                self.valueChanged.emit(
                    float(row + self._cache.dataYRange[0]),
                    float('nan'),
                    data[row])

    def _activeImageChangedSlot(self, previous, legend):
        """Handle Plot active image change.

        Resets side histograms cache
        """
        self._dirtyCache()
        self._updateHistograms()

    def setProfileWindowBehavior(self, behavior: Union[str, ProfileWindowBehavior]):
        """Set where profile widgets are displayed.

        :param ProfileWindowBehavior behavior:
        - 'popup': All profiles are displayed in pop-up windows
        - 'embedded': Horizontal, vertical and cross profiles are displayed in
          sides widgets, others are displayed in pop-up windows.
        """
        behavior = self.ProfileWindowBehavior.from_value(behavior)
        if behavior is not self.getProfileWindowBehavior():
            manager = self.__profile.getProfileManager()
            manager.clearProfile()
            manager.requestUpdateAllProfile()

            if behavior is self.ProfileWindowBehavior.EMBEDDED:
                horizontalProfileWindow = self._histoHPlot
                verticalProfileWindow = self._histoVPlot
            else:
                horizontalProfileWindow = None
                verticalProfileWindow = None

            manager.setSpecializedProfileWindow(
                rois.ProfileImageHorizontalLineROI, horizontalProfileWindow
            )
            manager.setSpecializedProfileWindow(
                rois.ProfileImageVerticalLineROI, verticalProfileWindow
            )
            self.__profileWindowBehavior = behavior

    def getProfileWindowBehavior(self) -> ProfileWindowBehavior:
        """Returns current profile display behavior.

        See :meth:`setProfileWindowBehavior` and :class:`ProfileWindowBehavior`
        """
        return self.__profileWindowBehavior

    def getProfileToolBar(self):
        """"Returns profile tools attached to this plot.

        :rtype: silx.gui.plot.PlotTools.ProfileToolBar
        """
        return self.__profile

    @property
    @deprecated(replacement="getProfileToolBar()")
    def profile(self):
        return self.getProfileToolBar()

    def getHistogram(self, axis):
        """Return the histogram and corresponding row or column extent.

        The returned value when an histogram is available is a dict with keys:

        - 'data': numpy array of the histogram values.
        - 'extent': (start, end) row or column index.
          end index is not included in the histogram.

        :param str axis: 'x' for horizontal, 'y' for vertical
        :return: The histogram and its extent as a dict or None.
        :rtype: dict
        """
        assert axis in ('x', 'y')
        if self._cache is None:
            return None
        else:
            if axis == 'x':
                return dict(
                    data=numpy.array(self._cache.histoH, copy=True),
                    extent=self._cache.dataXRange)
            else:
                return dict(
                    data=numpy.array(self._cache.histoV, copy=True),
                    extent=(self._cache.dataYRange))

    def radarView(self):
        """Get the lower right radarView widget."""
        return self._radarView

    def setRadarView(self, radarView):
        """Change the lower right radarView widget.

        :param RadarView radarView: Widget subclassing RadarView to replace
                                    the lower right corner widget.
        """
        self._radarView = radarView
        self._radarView.setPlotWidget(self)
        self.centralWidget().layout().addWidget(self._radarView, 1, 1)

    # High-level API

    def getColormap(self):
        """Get the default colormap description.

        :return: A description of the current colormap.
                 See :meth:`setColormap` for details.
        :rtype: dict
        """
        return self.getDefaultColormap()

    def setColormap(self, colormap=None, normalization=None,
                    autoscale=None, vmin=None, vmax=None, colors=None):
        """Set the default colormap and update active image.

        Parameters that are not provided are taken from the current colormap.

        The colormap parameter can also be a dict with the following keys:

        - *name*: string. The colormap to use:
          'gray', 'reversed gray', 'temperature', 'red', 'green', 'blue'.
        - *normalization*: string. The mapping to use for the colormap:
          either 'linear' or 'log'.
        - *autoscale*: bool. Whether to use autoscale (True)
          or range provided by keys 'vmin' and 'vmax' (False).
        - *vmin*: float. The minimum value of the range to use if 'autoscale'
          is False.
        - *vmax*: float. The maximum value of the range to use if 'autoscale'
          is False.
        - *colors*: optional. Nx3 or Nx4 array of float in [0, 1] or uint8.
                    List of RGB or RGBA colors to use (only if name is None)

        :param colormap: Name of the colormap in
            'gray', 'reversed gray', 'temperature', 'red', 'green', 'blue'.
            Or the description of the colormap as a dict.
        :type colormap: dict or str.
        :param str normalization: Colormap mapping: 'linear' or 'log'.
        :param bool autoscale: Whether to use autoscale (True)
                               or [vmin, vmax] range (False).
        :param float vmin: The minimum value of the range to use if
                           'autoscale' is False.
        :param float vmax: The maximum value of the range to use if
                           'autoscale' is False.
        :param numpy.ndarray colors: Only used if name is None.
            Custom colormap colors as Nx3 or Nx4 RGB or RGBA arrays
        """
        cmap = self.getDefaultColormap()

        if isinstance(colormap, Colormap):
            # Replace colormap
            cmap = colormap

            self.setDefaultColormap(cmap)

            # Update active image colormap
            activeImage = self.getActiveImage()
            if isinstance(activeImage, items.ColormapMixIn):
                activeImage.setColormap(cmap)

        elif isinstance(colormap, dict):
            # Support colormap parameter as a dict
            assert normalization is None
            assert autoscale is None
            assert vmin is None
            assert vmax is None
            assert colors is None
            cmap._setFromDict(colormap)

        else:
            if colormap is not None:
                cmap.setName(colormap)
            if normalization is not None:
                cmap.setNormalization(normalization)
            if autoscale:
                cmap.setVRange(None, None)
            else:
                if vmin is not None:
                    cmap.setVMin(vmin)
                if vmax is not None:
                    cmap.setVMax(vmax)
            if colors is not None:
                cmap.setColormapLUT(colors)

        cursorColor = cursorColorForColormap(cmap.getName())
        self.setInteractiveMode('zoom', color=cursorColor)

    def setImage(self, image, origin=(0, 0), scale=(1., 1.),
                 copy=True, reset=None, resetzoom=True):
        """Set the image to display.

        :param image: A 2D array representing the image or None to empty plot.
        :type image: numpy.ndarray-like with 2 dimensions or None.
        :param origin: The (x, y) position of the origin of the image.
                       Default: (0, 0).
                       The origin is the lower left corner of the image when
                       the Y axis is not inverted.
        :type origin: Tuple of 2 floats: (origin x, origin y).
        :param scale: The scale factor to apply to the image on X and Y axes.
                      Default: (1, 1).
                      It is the size of a pixel in the coordinates of the axes.
                      Scales must be positive numbers.
        :type scale: Tuple of 2 floats: (scale x, scale y).
        :param bool copy: Whether to copy image data (default) or not.
        :param bool reset: Deprecated. Alias for `resetzoom`.
        :param bool resetzoom: Whether to reset zoom and ROI (default) or not.
        """
        self._dirtyCache()

        if reset is not None:
            resetzoom = reset

        assert len(origin) == 2
        assert len(scale) == 2
        assert scale[0] > 0
        assert scale[1] > 0

        if image is None:
            self.remove(self._imageLegend, kind='image')
            return

        data = numpy.array(image, order='C', copy=copy)
        if data.size == 0:
            self.remove(self._imageLegend, kind='image')
            return

        assert data.ndim == 2 or (data.ndim == 3 and data.shape[2] in (3, 4))

        aggregation = self.getAggregationModeAction().getAggregationMode()
        if data.ndim != 2 and aggregation is not None:
            # RGB/A with aggregation is not supported
            aggregation = items.ImageDataAggregated.Aggregation.NONE

        if aggregation is items.ImageDataAggregated.Aggregation.NONE:
            self.addImage(data,
                          legend=self._imageLegend,
                          origin=origin, scale=scale,
                          colormap=self.getColormap(),
                          resetzoom=False)
        else:
            item = self._getItem("image", self._imageLegend)
            if isinstance(item, items.ImageDataAggregated):
                item.setData(data)
                item.setOrigin(origin)
                item.setScale(scale)
            else:
                if isinstance(item, items.ImageDataAggregated):
                    imageItem = item
                    wasCreated = False
                else:
                    if item is not None:
                        self.removeImage(self._imageLegend)
                    imageItem = items.ImageDataAggregated()
                    imageItem.setName(self._imageLegend)
                    imageItem.setColormap(self.getColormap())
                    wasCreated = True
                imageItem.setData(data)
                imageItem.setOrigin(origin)
                imageItem.setScale(scale)
                imageItem.setAggregationMode(aggregation)
                if wasCreated:
                    self.addItem(imageItem)

        self.setActiveImage(self._imageLegend)
        self._updateHistograms()
        if resetzoom:
            self.resetZoom()


# ImageViewMainWindow #########################################################

class ImageViewMainWindow(ImageView):
    """:class:`ImageView` with additional toolbars

    Adds extra toolbar and a status bar to :class:`ImageView`.
    """
    def __init__(self, parent=None, backend=None):
        self._dataInfo = None
        super(ImageViewMainWindow, self).__init__(parent, backend)
        self.setWindowFlags(qt.Qt.Window)

        self.getXAxis().setLabel('X')
        self.getYAxis().setLabel('Y')
        self.setGraphTitle('Image')

        # Add toolbars and status bar
        self.addToolBar(qt.Qt.BottomToolBarArea, LimitsToolBar(plot=self))

        menu = self.menuBar().addMenu('File')
        menu.addAction(self.getOutputToolBar().getSaveAction())
        menu.addAction(self.getOutputToolBar().getPrintAction())
        menu.addSeparator()
        action = menu.addAction('Quit')
        action.triggered[bool].connect(qt.QApplication.instance().quit)

        menu = self.menuBar().addMenu('Edit')
        menu.addAction(self.getOutputToolBar().getCopyAction())
        menu.addSeparator()
        menu.addAction(self.getResetZoomAction())
        menu.addAction(self.getColormapAction())
        menu.addAction(actions.control.KeepAspectRatioAction(self, self))
        menu.addAction(actions.control.YAxisInvertedAction(self, self))
        menu.addAction(self.getShowSideHistogramsAction())

        self.__profileMenu = self.menuBar().addMenu('Profile')
        self.__updateProfileMenu()

        # Connect to ImageView's signal
        self.valueChanged.connect(self._statusBarSlot)

    def __updateProfileMenu(self):
        """Update actions available in 'Profile' menu"""
        profile = self.getProfileToolBar()
        self.__profileMenu.clear()
        self.__profileMenu.addAction(profile.hLineAction)
        self.__profileMenu.addAction(profile.vLineAction)
        self.__profileMenu.addAction(profile.crossAction)
        self.__profileMenu.addAction(profile.lineAction)
        self.__profileMenu.addAction(profile.clearAction)

    def _formatValueToString(self, value):
        try:
            if isinstance(value, numpy.ndarray):
                if len(value) == 4:
                    return "RGBA: %.3g, %.3g, %.3g, %.3g" % (value[0], value[1], value[2], value[3])
                elif len(value) == 3:
                    return "RGB: %.3g, %.3g, %.3g" % (value[0], value[1], value[2])
            else:
                return "Value: %g" % value
        except Exception:
            _logger.error("Error while formatting pixel value", exc_info=True)
            pass
        return "Value: %s" % value

    def _statusBarSlot(self, row, column, value):
        """Update status bar with coordinates/value from plots."""
        if numpy.isnan(row):
            msg = 'Column: %d, Sum: %g' % (int(column), value)
        elif numpy.isnan(column):
            msg = 'Row: %d, Sum: %g' % (int(row), value)
        else:
            msg_value = self._formatValueToString(value)
            msg = 'Position: (%d, %d), %s' % (int(row), int(column), msg_value)
        if self._dataInfo is not None:
            msg = self._dataInfo + ', ' + msg

        self.statusBar().showMessage(msg)

    @docstring(ImageView)
    def setProfileWindowBehavior(self, behavior: str):
        super().setProfileWindowBehavior(behavior)
        self.__updateProfileMenu()

    @docstring(ImageView)
    def setImage(self, image, *args, **kwargs):
        if hasattr(image, 'dtype') and hasattr(image, 'shape'):
            assert image.ndim == 2 or (image.ndim == 3 and image.shape[2] in (3, 4))
            height, width = image.shape[0:2]
            dataInfo = 'Data: %dx%d (%s)' % (width, height, str(image.dtype))
        else:
            dataInfo = None

        if self._dataInfo != dataInfo:
            self._dataInfo = dataInfo
            self.statusBar().showMessage(self._dataInfo)

        # Set the new image in ImageView widget
        super(ImageViewMainWindow, self).setImage(image, *args, **kwargs)



# /*##########################################################################
#
# Copyright (c) 2015-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides median filter function for 1D and 2D arrays.
"""

__authors__ = ["H. Payno", "J. Kieffer"]
__license__ = "MIT"
__date__ = "02/05/2017"


from cython.parallel import prange
cimport cython
cimport silx.math.medianfilter.median_filter as median_filter
import numpy
cimport numpy as cnumpy
from libcpp cimport bool

import numbers

ctypedef unsigned long uint64
ctypedef unsigned int uint32
ctypedef unsigned short uint16


MODES = {'nearest': 0, 'reflect': 1, 'mirror': 2, 'shrink': 3, 'constant': 4}


def medfilt1d(data,
              kernel_size=3,
              bool conditional=False,
              mode='nearest',
              cval=0):
    """Function computing the median filter of the given input.

    Behavior at boundaries: the algorithm is reducing the size of the
    window/kernel for pixels at boundaries (there is no mirroring).

    Not-a-Number (NaN) float values are ignored.
    If the window only contains NaNs, it evaluates to NaN.

    In event of an even number of valid values in the window (either
    because of NaN values or on image border in shrink mode),
    the highest of the 2 central sorted values is taken.

    :param numpy.ndarray data: the array for which we want to apply
        the median filter. Should be 1d.
    :param kernel_size: the dimension of the kernel.
    :type kernel_size: int
    :param bool conditional: True if we want to apply a conditional median
        filtering.
    :param str mode: the algorithm used to determine how values at borders
        are determined: 'nearest', 'reflect', 'mirror', 'shrink', 'constant'
    :param cval: Value used outside borders in 'constant' mode

    :returns: the array with the median value for each pixel.
    """
    return medfilt(data, kernel_size, conditional, mode, cval)


def medfilt2d(image,
              kernel_size=3,
              bool conditional=False,
              mode='nearest',
              cval=0):
    """Function computing the median filter of the given input.
    Behavior at boundaries: the algorithm is reducing the size of the
    window/kernel for pixels at boundaries (there is no mirroring).

    Not-a-Number (NaN) float values are ignored.
    If the window only contains NaNs, it evaluates to NaN.

    In event of an even number of valid values in the window (either
    because of NaN values or on image border in shrink mode),
    the highest of the 2 central sorted values is taken.

    :param numpy.ndarray data: the array for which we want to apply
        the median filter. Should be 2d.
    :param kernel_size: the dimension of the kernel.
    :type kernel_size: For 1D should be an int for 2D should be a tuple or
        a list of (kernel_height, kernel_width)
    :param bool conditional: True if we want to apply a conditional median
        filtering.
    :param str mode: the algorithm used to determine how values at borders
        are determined: 'nearest', 'reflect', 'mirror', 'shrink', 'constant'
    :param cval: Value used outside borders in 'constant' mode

    :returns: the array with the median value for each pixel.
    """
    return medfilt(image, kernel_size, conditional, mode, cval)


def medfilt(data,
            kernel_size=3,
            bool conditional=False,
            mode='nearest',
            cval=0):
    """Function computing the median filter of the given input.
    Behavior at boundaries: the algorithm is reducing the size of the
    window/kernel for pixels at boundaries (there is no mirroring).

    Not-a-Number (NaN) float values are ignored.
    If the window only contains NaNs, it evaluates to NaN.

    In event of an even number of valid values in the window (either
    because of NaN values or on image border in shrink mode),
    the highest of the 2 central sorted values is taken.

    :param numpy.ndarray data: the array for which we want to apply
        the median filter. Should be 1d or 2d.
    :param kernel_size: the dimension of the kernel.
    :type kernel_size: For 1D should be an int for 2D should be a tuple or
        a list of (kernel_height, kernel_width)
    :param bool conditional: True if we want to apply a conditional median
        filtering.
    :param str mode: the algorithm used to determine how values at borders
        are determined: 'nearest', 'reflect', 'mirror', 'shrink', 'constant'
    :param cval: Value used outside borders in 'constant' mode

    :returns: the array with the median value for each pixel.
    """
    if mode not in MODES:
        err = 'Requested mode %s is unknown.' % mode
        raise ValueError(err)

    if data.ndim > 2:
        raise ValueError(
            "Invalid data shape. Dimension of the array should be 1 or 2")

    # Handle case of scalar kernel size
    if isinstance(kernel_size, numbers.Integral):
        kernel_size = [kernel_size] * data.ndim

    assert len(kernel_size) == data.ndim

    # Convert 1D arrays to 2D
    reshaped = False
    if len(data.shape) == 1:
        data = data.reshape(1, data.shape[0])
        kernel_size = [1, kernel_size[0]]
        reshaped = True

    # simple median filter apply into a 2D buffer
    output_buffer = numpy.zeros_like(data)
    check(data, output_buffer)

    ker_dim = numpy.array(kernel_size, dtype=numpy.int32)

    if data.dtype == numpy.float64:
        medfilterfc = _median_filter_float64
    elif data.dtype == numpy.float32:
        medfilterfc = _median_filter_float32
    elif data.dtype == numpy.int64:
        medfilterfc = _median_filter_int64
    elif data.dtype == numpy.uint64:
        medfilterfc = _median_filter_uint64
    elif data.dtype == numpy.int32:
        medfilterfc = _median_filter_int32
    elif data.dtype == numpy.uint32:
        medfilterfc = _median_filter_uint32
    elif data.dtype == numpy.int16:
        medfilterfc = _median_filter_int16
    elif data.dtype == numpy.uint16:
        medfilterfc = _median_filter_uint16
    else:
        raise ValueError("%s type is not managed by the median filter" % data.dtype)

    medfilterfc(input_buffer=data,
                output_buffer=output_buffer,
                kernel_size=ker_dim,
                conditional=conditional,
                mode=MODES[mode],
                cval=cval)

    if reshaped:
        output_buffer.shape = -1  # Convert to 1D array

    return output_buffer


def check(input_buffer, output_buffer):
    """Simple check on the two buffers to make sure we can apply the median filter
    """
    if (input_buffer.flags['C_CONTIGUOUS'] is False):
        raise ValueError('<input_buffer> must be a C_CONTIGUOUS numpy array.')

    if (output_buffer.flags['C_CONTIGUOUS'] is False):
        raise ValueError('<output_buffer> must be a C_CONTIGUOUS numpy array.')

    if not (len(input_buffer.shape) <= 2):
        raise ValueError('<input_buffer> dimension must mo higher than 2.')

    if not (len(output_buffer.shape) <= 2):
        raise ValueError('<output_buffer> dimension must mo higher than 2.')

    if not(input_buffer.dtype == output_buffer.dtype):
        raise ValueError('input buffer and output_buffer must be of the same type')

    if not (input_buffer.shape == output_buffer.shape):
        raise ValueError('input buffer and output_buffer must be of the same dimension and same dimension')


######### implementations of the include/median_filter.hpp function ############
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def reflect(int index, int length_max):
    """find the correct index into [0, length_max-1] for index in reflect mode

    :param int index: the index to move into [0, length_max-1] in reflect mode
    :param int length_max: the higher bound limit
    """
    return median_filter.reflect(index, length_max)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def mirror(int index, int length_max):
    """find the correct index into [0, length_max-1] for index in mirror mode

    :param int index: the index to move into [0, length_max-1] in mirror mode
    :param int length_max: the higher bound limit
    """
    return median_filter.mirror(index, length_max)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def _median_filter_float32(float[:, ::1] input_buffer not None,
                           float[:, ::1] output_buffer not None,
                           cnumpy.int32_t[::1] kernel_size not None,
                           bool conditional,
                           int mode,
                           float cval):

    cdef:
        int y = 0
        int image_dim = input_buffer.shape[1] - 1
        int[2] buffer_shape
    buffer_shape[0] = input_buffer.shape[0]
    buffer_shape[1] = input_buffer.shape[1]

    for y in prange(input_buffer.shape[0], nogil=True):
            median_filter.median_filter[float](<float*> & input_buffer[0,0], 
                                               <float*> & output_buffer[0,0], 
                                               <int*>& kernel_size[0],
                                               <int*>buffer_shape,
                                               y,
                                               0,
                                               image_dim,
                                               conditional,
                                               mode,
                                               cval)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def _median_filter_float64(double[:, ::1] input_buffer not None,
                           double[:, ::1] output_buffer not None,
                           cnumpy.int32_t[::1] kernel_size not None,
                           bool conditional,
                           int mode,
                           double cval):

    cdef:
        int y = 0
        int image_dim = input_buffer.shape[1] - 1
        int[2] buffer_shape
    buffer_shape[0] = input_buffer.shape[0]
    buffer_shape[1] = input_buffer.shape[1]

    for y in prange(input_buffer.shape[0], nogil=True):
            median_filter.median_filter[double](<double*> & input_buffer[0, 0], 
                                                <double*> & output_buffer[0, 0], 
                                                <int*>&kernel_size[0],
                                                <int*>buffer_shape,
                                                y,
                                                0,
                                                image_dim,
                                                conditional,
                                                mode,
                                                cval)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def _median_filter_int64(cnumpy.int64_t[:, ::1] input_buffer not None,
                         cnumpy.int64_t[:, ::1] output_buffer not None,
                         cnumpy.int32_t[::1] kernel_size not None,
                         bool conditional,
                         int mode,
                         cnumpy.int64_t cval):

    cdef:
        int y = 0
        int image_dim = input_buffer.shape[1] - 1
        int[2] buffer_shape
    buffer_shape[0] = input_buffer.shape[0]
    buffer_shape[1] = input_buffer.shape[1]

    for y in prange(input_buffer.shape[0], nogil=True):
            median_filter.median_filter[long](<long*> & input_buffer[0,0], 
                                              <long*>  & output_buffer[0, 0], 
                                              <int*>&kernel_size[0],
                                              <int*>buffer_shape,
                                              y,
                                              0,
                                              image_dim,
                                              conditional,
                                              mode,
                                              cval)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def _median_filter_uint64(cnumpy.uint64_t[:, ::1] input_buffer not None,
                          cnumpy.uint64_t[:, ::1] output_buffer not None,
                          cnumpy.int32_t[::1] kernel_size not None,
                          bool conditional,
                          int mode,
                          cnumpy.uint64_t cval):

    cdef: 
        int y = 0
        int image_dim = input_buffer.shape[1] - 1
        int[2] buffer_shape
    buffer_shape[0] = input_buffer.shape[0]
    buffer_shape[1] = input_buffer.shape[1]

    for y in prange(input_buffer.shape[0], nogil=True):
            median_filter.median_filter[uint64](<uint64*> & input_buffer[0,0], 
                                                <uint64*> & output_buffer[0, 0],
                                                <int*>&kernel_size[0],
                                                <int*>buffer_shape,
                                                y,
                                                0,
                                                image_dim,
                                                conditional,
                                                mode,
                                                cval)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def _median_filter_int32(cnumpy.int32_t[:, ::1] input_buffer not None,
                         cnumpy.int32_t[:, ::1] output_buffer not None,
                         cnumpy.int32_t[::1] kernel_size not None,
                         bool conditional,
                         int mode,
                         cnumpy.int32_t cval):

    cdef:
        int y = 0
        int image_dim = input_buffer.shape[1] - 1
        int[2] buffer_shape
    buffer_shape[0] = input_buffer.shape[0]
    buffer_shape[1] = input_buffer.shape[1]

    for y in prange(input_buffer.shape[0], nogil=True):
            median_filter.median_filter[int](<int*> & input_buffer[0,0], 
                                             <int*>  & output_buffer[0, 0],
                                             <int*>&kernel_size[0],
                                             <int*>buffer_shape,
                                             y,
                                             0,
                                             image_dim,
                                             conditional,
                                             mode,
                                             cval)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def _median_filter_uint32(cnumpy.uint32_t[:, ::1] input_buffer not None,
                          cnumpy.uint32_t[:, ::1] output_buffer not None,
                          cnumpy.int32_t[::1] kernel_size not None,
                          bool conditional,
                          int mode,
                          cnumpy.uint32_t cval):

    cdef:
        int y = 0
        int image_dim = input_buffer.shape[1] - 1
        int[2] buffer_shape
    buffer_shape[0] = input_buffer.shape[0]
    buffer_shape[1] = input_buffer.shape[1]

    for y in prange(input_buffer.shape[0], nogil=True):
            median_filter.median_filter[uint32](<uint32*> & input_buffer[0,0], 
                                                <uint32*>  & output_buffer[0, 0],
                                                <int*>&kernel_size[0],
                                                <int*>buffer_shape,
                                                y,
                                                0,
                                                image_dim,
                                                conditional,
                                                mode,
                                                cval)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def _median_filter_int16(cnumpy.int16_t[:, ::1] input_buffer not None,
                         cnumpy.int16_t[:, ::1] output_buffer not None,
                         cnumpy.int32_t[::1] kernel_size not None,
                         bool conditional,
                         int mode,
                         cnumpy.int16_t cval):

    cdef:
        int y = 0
        int image_dim = input_buffer.shape[1] - 1
        int[2] buffer_shape
    buffer_shape[0] = input_buffer.shape[0]
    buffer_shape[1] = input_buffer.shape[1]

    for y in prange(input_buffer.shape[0], nogil=True):
            median_filter.median_filter[short](<short*> & input_buffer[0,0], 
                                               <short*>  & output_buffer[0, 0],
                                               <int*>&kernel_size[0],
                                               <int*>buffer_shape,
                                               y,
                                               0,
                                               image_dim,
                                               conditional,
                                               mode,
                                               cval)


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def _median_filter_uint16(
      cnumpy.uint16_t[:, ::1] input_buffer not None,
      cnumpy.uint16_t[:, ::1] output_buffer not None,
      cnumpy.int32_t[::1] kernel_size not None,
      bool conditional,
      int mode,
      cnumpy.uint16_t cval):

    cdef:
        int y = 0
        int image_dim = input_buffer.shape[1] - 1
        int[2] buffer_shape, 
    buffer_shape[0] = input_buffer.shape[0]
    buffer_shape[1] = input_buffer.shape[1]

    for y in prange(input_buffer.shape[0], nogil=True):
            median_filter.median_filter[uint16](<uint16*> & input_buffer[0, 0],
                                                <uint16*> & output_buffer[0, 0],
                                                <int*>&kernel_size[0],
                                                <int*>buffer_shape,
                                                y,
                                                0,
                                                image_dim,
                                                conditional,
                                                mode,
                                                cval)



# /*##########################################################################
#
# Copyright (c) 2017-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
This module provides base classes to implement models for 3D scene content.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "11/01/2018"


import collections
import weakref

from ....utils.weakref import WeakMethodProxy
from ... import qt


class BaseRow(qt.QObject):
    """Base class for rows of the tree model.

    The root node parent MUST be set to the QAbstractItemModel it belongs to.
    By default item is enabled.

    :param children: Iterable of BaseRow to start with (not signaled)
    """

    def __init__(self, children=()):
        self.__modelRef = None
        self.__parentRef = None
        super(BaseRow, self).__init__()
        self.__children = []
        for row in children:
            assert isinstance(row, BaseRow)
            row.setParent(self)
            self.__children.append(row)
        self.__flags = collections.defaultdict(lambda: qt.Qt.ItemIsEnabled)
        self.__tooltip = None

    def setParent(self, parent):
        """Override :meth:`QObject.setParent` to cache model and parent"""
        self.__parentRef = None if parent is None else weakref.ref(parent)

        if isinstance(parent, qt.QAbstractItemModel):
            model = parent
        elif isinstance(parent, BaseRow):
            model = parent.model()
        else:
            model = None

        self._updateModel(model)

        super(BaseRow, self).setParent(parent)

    def parent(self):
        """Override :meth:`QObject.setParent` to use cached parent

        :rtype: Union[QObject, None]"""
        return self.__parentRef() if self.__parentRef is not None else None

    def _updateModel(self, model):
        """Update the model this row belongs to"""
        if model != self.model():
            self.__modelRef = weakref.ref(model) if model is not None else None
            for child in self.children():
                child._updateModel(model)

    def model(self):
        """Return the model this node belongs to or None if not in a model.

        :rtype: Union[QAbstractItemModel, None]
        """
        return self.__modelRef() if self.__modelRef is not None else None

    def index(self, column=0):
        """Return corresponding index in the model or None if not in a model.

        :param int column: The column to make the index for
        :rtype: Union[QModelIndex, None]
        """
        parent = self.parent()
        model = self.model()

        if model is None:  # Not in a model
            return None
        elif parent is model:  # Root node
            return qt.QModelIndex()
        else:
            index = parent.index()
            row = parent.children().index(self)
            return model.index(row, column, index)

    def columnCount(self):
        """Returns number of columns (default: 2)

        :rtype: int
        """
        return 2

    def children(self):
        """Returns the list of children nodes

        :rtype: tuple of Node
        """
        return tuple(self.__children)

    def rowCount(self):
        """Returns number of rows

        :rtype: int
        """
        return len(self.__children)

    def addRow(self, row, index=None):
        """Add a node to the children

        :param BaseRow row: The node to add
        :param int index: The index at which to insert it or
                          None to append
        """
        if index is None:
            index = self.rowCount()
        assert index <= self.rowCount()

        model = self.model()

        if model is not None:
            parent = self.index()
            model.beginInsertRows(parent, index, index)

        self.__children.insert(index, row)
        row.setParent(self)

        if model is not None:
            model.endInsertRows()

    def removeRow(self, row):
        """Remove a row from the children list.

        It removes either a node or a row index.

        :param row: BaseRow object or index of row to remove
        :type row: Union[BaseRow, int]
        """
        if isinstance(row, BaseRow):
            row = self.__children.index(row)
        else:
            row = int(row)
        assert row < self.rowCount()

        model = self.model()

        if model is not None:
            index = self.index()
            model.beginRemoveRows(index, row, row)

        node = self.__children.pop(row)
        node.setParent(None)

        if model is not None:
            model.endRemoveRows()

    def data(self, column, role):
        """Returns data for given column and role

        :param int column: Column index for this row
        :param int role: The role to get
        :return: Corresponding data (Default: None)
        """
        if role == qt.Qt.ToolTipRole and self.__tooltip is not None:
            return self.__tooltip
        else:
            return None

    def setData(self, column, value, role):
        """Set data for given column and role

        :param int column: Column index for this row
        :param value: The data to set
        :param int role: The role to set
        :return: True on success, False on failure
        :rtype: bool
        """
        return False

    def setToolTip(self, tooltip):
        """Set the tooltip of the whole row.

        If None there is no tooltip.

        :param Union[str, None] tooltip:
        """
        self.__tooltip = tooltip

    def setFlags(self, flags, column=None):
        """Set the static flags to return.

        Default is ItemIsEnabled for all columns.

        :param int column: The column for which to set the flags
        :param flags: Item flags
        """
        if column is None:
            self.__flags = collections.defaultdict(lambda: flags)
        else:
            self.__flags[column] = flags

    def flags(self, column):
        """Returns flags for given column

        :rtype: int
        """
        return self.__flags[column]


class StaticRow(BaseRow):
    """Row with static data.

    :param tuple display: List of data for DisplayRole for each column
    :param dict roles: Optional mapping of roles to list of data.
    :param children: Iterable of BaseRow to start with (not signaled)
    """

    def __init__(self, display=('', None), roles=None, children=()):
        super(StaticRow, self).__init__(children)
        self._dataByRoles = {} if roles is None else roles
        self._dataByRoles[qt.Qt.DisplayRole] = display

    def data(self, column, role):
        if role in self._dataByRoles:
            data = self._dataByRoles[role]
            if column < len(data):
                return data[column]
        return super(StaticRow, self).data(column, role)

    def columnCount(self):
        return len(self._dataByRoles[qt.Qt.DisplayRole])


class ProxyRow(BaseRow):
    """Provides a node to proxy a data accessible through functions.

    Warning: Only weak reference are kept on fget and fset.

    :param str name: The name of this node
    :param callable fget: A callable returning the data
    :param callable fset:
        An optional callable setting the data with data as a single argument.
    :param notify:
        An optional signal emitted when data has changed.
    :param callable toModelData:
        An optional callable to convert from fget
        callable to data returned by the model.
    :param callable fromModelData:
        An optional callable converting data provided to the model to
        data for fset.
    :param editorHint: Data to provide as UserRole for editor selection/setup
    """

    def __init__(self,
                 name='',
                 fget=None,
                 fset=None,
                 notify=None,
                 toModelData=None,
                 fromModelData=None,
                 editorHint=None):

        super(ProxyRow, self).__init__()
        self.__name = name
        self.__editorHint = editorHint

        assert fget is not None
        self._fget = WeakMethodProxy(fget)
        self._fset = WeakMethodProxy(fset) if fset is not None else None
        if fset is not None:
            self.setFlags(qt.Qt.ItemIsEnabled | qt.Qt.ItemIsEditable, 1)
        self._toModelData = toModelData
        self._fromModelData = fromModelData

        if notify is not None:
            notify.connect(self._notified)  # TODO support sigItemChanged flags

    def _notified(self, *args, **kwargs):
        """Send update to the model upon signal notifications"""
        index = self.index(column=1)
        model = self.model()
        if model is not None:
            model.dataChanged.emit(index, index)

    def data(self, column, role):
        if column == 0:
            if role == qt.Qt.DisplayRole:
                return self.__name

        elif column == 1:
            if role == qt.Qt.UserRole:  # EditorHint
                return self.__editorHint
            elif role == qt.Qt.DisplayRole or (role == qt.Qt.EditRole and
                                               self._fset is not None):
                data = self._fget()
                if self._toModelData is not None:
                    data = self._toModelData(data)
                return data

        return super(ProxyRow, self).data(column, role)

    def setData(self, column, value, role):
        if role == qt.Qt.EditRole and self._fset is not None:
            if self._fromModelData is not None:
                value = self._fromModelData(value)
            self._fset(value)
            return True

        return super(ProxyRow, self).setData(column, value, role)


class ColorProxyRow(ProxyRow):
    """Provides a proxy to a QColor property.

    The color is returned through the decorative role.

    See :class:`ProxyRow`
    """

    def data(self, column, role):
        if column == 1:  # Show color as decoration, not text
            if role == qt.Qt.DisplayRole:
                return None
            if role == qt.Qt.DecorationRole:
                role = qt.Qt.DisplayRole
        return super(ColorProxyRow, self).data(column, role)


class AngleDegreeRow(ProxyRow):
    """ProxyRow patching display of column 1 to add degree symbol

    See :class:`ProxyRow`
    """

    def __init__(self, *args, **kwargs):
        super(AngleDegreeRow, self).__init__(*args, **kwargs)

    def data(self, column, role):
        if column == 1 and role == qt.Qt.DisplayRole:
            return u'%g°' % super(AngleDegreeRow, self).data(column, role)
        else:
            return super(AngleDegreeRow, self).data(column, role)



# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests that the different implementation of opencl (cpp, opencl) are
    accessible
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/05/2017"

import unittest
from silx.image import medianfilter
import numpy

from silx.opencl.common import ocl


class TestMedianFilterEngines(unittest.TestCase):
    """Make sure we have access to all the different implementation of
    median filter from image medfilt"""


    IMG = numpy.arange(10000.).reshape(100, 100)

    KERNEL = (1, 1)

    def testCppMedFilt2d(self):
        """test cpp engine for medfilt2d"""
        res = medianfilter.medfilt2d(
            image=TestMedianFilterEngines.IMG,
            kernel_size=TestMedianFilterEngines.KERNEL,
            engine='cpp')
        self.assertTrue(numpy.array_equal(res, TestMedianFilterEngines.IMG))

    @unittest.skipUnless(ocl, "PyOpenCl is missing")
    def testOpenCLMedFilt2d(self):
        """test cpp engine for medfilt2d"""
        res = medianfilter.medfilt2d(
            image=TestMedianFilterEngines.IMG,
            kernel_size=TestMedianFilterEngines.KERNEL,
            engine='opencl')
        self.assertTrue(numpy.array_equal(res, TestMedianFilterEngines.IMG))



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""silx.gui.plot Plot2D Qt designer plugin."""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "30/05/2016"


from silx.gui import icons, qt

if qt.BINDING == 'PyQt5':
    from PyQt5 import QtDesigner
else:
    raise RuntimeError("Unsupport Qt BINDING: %s" % qt.BINDING)

from silx.gui.plot import Plot2D


class Plot2DPlugin(QtDesigner.QPyDesignerCustomWidgetPlugin):

    def __init__(self, parent=None):
        super(Plot2DPlugin, self).__init__(parent)
        self.initialized = False

    def initialize(self, core):
        if self.initialized:
            return

        self.initialized = True

    def isInitialized(self):
        return self.initialized

    def createWidget(self, parent):
        plot = Plot2D(parent=parent)
        plot.setAutoReplot(False)
        return plot

    def name(self):
        return "Plot2D"

    def group(self):
        return "silx"

    def icon(self):
        return icons.getQIcon('plot-window-image')

    def toolTip(self):
        return ""

    def whatsThis(self):
        return ""

    def isContainer(self):
        return False

    def includeFile(self):
        return "silx.gui.plot.PlotWindow"



# /*##########################################################################
#
# Copyright (c) 2015-2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

from libcpp.vector cimport vector as std_vector
from libcpp cimport bool

cdef extern from "mc.hpp":
    cdef cppclass MarchingCubes[FloatIn, FloatOut]:
        MarchingCubes(FloatIn level) except +
        void process(FloatIn * data,
                     unsigned int depth,
                     unsigned int height,
                     unsigned int width) except +
        void set_slice_size(unsigned int height,
                            unsigned int width)
        void process_slice(FloatIn * slice0,
                           FloatIn * slice1) except +
        void finish_process()
        void reset()

        unsigned int depth
        unsigned int height
        unsigned int width
        unsigned int sampling[3]
        FloatIn isolevel
        bool invert_normals
        std_vector[FloatOut] vertices
        std_vector[FloatOut] normals
        std_vector[unsigned int] indices
